/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 * evaluate.c
 * trace condition tree evaluation logic
 */

#include "fist.h"
#include "structures.h"
#include "tracer.h"
#include "namecache.h"
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif

#define TRUE  1
#define FALSE 0

/**
 * Evaluates the trace condition tree.
 * Evaluates whether this operation should be traced
 * given the appropriate inode and dentry. This is a recursive
 * call.
 * @param root   root of the trace condition tree.
 * @param op     Operation to be traced.
 * @param inode  Inode passed to the operation.
 * @param dentry Dentry passed to the operation.
 * @return boolean value indicating whether operation should be logged
 */
static int _evaluate(struct tracecond *root, op_t op, struct inode *inode, struct dentry *dentry)
{
    int result = TRUE;
    print_entry_location();

    while (root) {
    	root->tc_num_eval++;
    	switch (root->tc_type) {
    	case TC_TRUE: {
		result = TRUE;
		break;
    	}

    	case TC_FALSE: {
		result = FALSE;
		break;
    	}
    	case TC_OP: {
		ASSERT(op < OP_MAX);
		result = OP_IS_ENABLED(root->u.tc_map, op);
		break;
    	}
#ifdef TRACE_SYSCALL
    	case TC_SYSCALL:
		result = SYSCALL_IS_ENABLED(root->u.tc_map, current->sys_call);

		break;
#endif
    	case TC_CUID: {
		result = (root->u.tc_ival == current->euid);
		break;
    	}
    	case TC_CGID: {
		result = (root->u.tc_ival == current->egid);
		break;
    	}
    	case TC_EXT:
    	case TC_NAME: {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	  result = inode && lookup_fid(root->u.tc_name, inode->i_ino, &inode->i_dev);
	  if (!result && dentry && ((inode = dentry->d_inode) != NULL))
	    result = lookup_fid(root->u.tc_name, inode->i_ino, &inode->i_dev) != NULL;
#else
	  result = inode && lookup_fid(root->u.tc_name, inode->i_ino, &inode->i_sb->s_dev);
	  if (!result && dentry && ((inode = dentry->d_inode) != NULL))
	    result = lookup_fid(root->u.tc_name, inode->i_ino, &inode->i_sb->s_dev) != NULL;
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	  break;
	}
    	case TC_PID: {
		result = (root->u.tc_ival == current->pid);
		break;
    	}
    	case TC_PGRP: {
	  /*
	   * PGRP - Process Group not implemented starting in kernel 2.6
	   */
	        //result = (root->u.tc_ival == current->pgrp);
		break;
    	}
    	case TC_FUID:
	    printk("<0>BUG: TC_FUID not implemented!\n");
	    ASSERT(NULL);
	    break;
    	case TC_FGID:
	    printk("<0>BUG: TC_FGID not implemented!\n");
	    ASSERT(NULL);
	    break;
    	case TC_PNAME: {
	    result = (strncmp(root->u.tc_map, current->comm, 16) == 0);
	    //printk("%s, %s", root->u.tc_map, current->comm);
	    break;
	}
    	case TC_SID: {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	  result = (root->u.tc_ival == current->session);
#else
	  result = (root->u.tc_ival == current->signal->session);
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	    break;
    	}
    	case TC_TIME:
	    printk("<0>BUG: TC_TIME not implemented!\n");
	    ASSERT(NULL);
	    break;
    	case TC_INO: {

		result = inode && (root->u.tc_ival == inode->i_ino);
		if (!result && dentry && dentry->d_inode)
	    		result = (root->u.tc_ival == dentry->d_inode->i_ino);
		break;
    	}
    	default:
		/* should not happen */
		ASSERT(0);
    	}
	/*
	 * if the type is true or false, these nodes have to be leaf nodes
	 * if the type is not true or false, the node must necessarily not be a leaf node
	 *   and should also have both children. Hence, this obscure ASSERT...
	 */
	/*
	 * Removing this assertion as this doesn't work for TRUE and FALSE expressions.
	 * The DAG _must_ have a TRUE and a FALSE termination, which means that there will
	 * be DAGs like
	 *              TRUE
	 *             /    \
	 *          TRUE   FALSE
	 * This is sort of silly, but that's how the user space parser works.
	 */

	//ASSERT( ((root->tc_type == TC_TRUE || root->tc_type == TC_FALSE) && (root->tc_follow_true == NULL) && (root->tc_follow_true == NULL))
	//	|| (root->tc_type != TC_TRUE && root->tc_type != TC_FALSE && root->tc_follow_true != NULL && root->tc_follow_false != NULL));
	if (result) {
		root->tc_num_eval_true++;
		root = root->tc_follow_true;
	} else {
		root = root->tc_follow_false;
	}
    }
    print_exit_status(result);
    return result;
}


/*
 * TODO:
 * once recursion is removed from _evaluate
 * the two functions can be combined into one
 */

int evaluate(struct trace_destination *td, op_t op, struct inode *inode, struct dentry *dentry)
{
    if (td->td_state != TD_ENABLED)
	return FALSE;
    return _evaluate(td->td_root, op, inode, dentry);
}
