/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: file.c,v 1.37 2007/01/30 06:30:53 gopalas Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#ifdef FISTGEN
# include "fist_tracefs.h"
#endif /* FISTGEN */
#include "fist.h"
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/security.h>
#endif
#include "tracefs.h"
#include "tracer.h"
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif

/*******************
 * File Operations *
 *******************/


STATIC loff_t
tracefs_llseek(file_t *file, loff_t offset, int origin)
{
    loff_t err;
    file_t *hidden_file = NULL;
    void *context;

    print_entry_location();

    trace_llseek_pre(stoptr(file->f_dentry->d_inode->i_sb), file, offset, origin, &context);

    if(ftopd(file) != NULL)
	hidden_file = ftohf(file); /* CPW: Moved to after print_entry_location */

    fist_dprint(6, "tracefs_llseek: file=%x, offset=0x%x, origin=%d\n",
		(int) file, (int) offset, origin);

    /* always set hidden position to this one */
    hidden_file->f_pos = file->f_pos;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    if (file->f_reada) { /* update readahead information if needed */
	hidden_file->f_reada = file->f_reada;
	hidden_file->f_ramax = file->f_ramax;
	hidden_file->f_raend = file->f_raend;
	hidden_file->f_ralen = file->f_ralen;
	hidden_file->f_rawin = file->f_rawin;
    }
    ASSERT(hidden_file->f_reada == file->f_reada);
    ASSERT(hidden_file->f_ramax == file->f_ramax);
    ASSERT(hidden_file->f_raend == file->f_raend);
    ASSERT(hidden_file->f_ralen == file->f_ralen);
    ASSERT(hidden_file->f_rawin == file->f_rawin);
#else
    memcpy(&hidden_file->f_ra,&(file->f_ra),sizeof(struct file_ra_state));
#endif

    if (hidden_file->f_op && hidden_file->f_op->llseek)
	err = hidden_file->f_op->llseek(hidden_file, offset, origin);
    else
	err = generic_file_llseek(hidden_file, offset, origin);

    if (err < 0)
	goto out;

    if (err != file->f_pos) {
	file->f_pos = err;
	// ION maybe this?
	// 	file->f_pos = hidden_file->f_pos;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	file->f_reada = 0; //Turns off read ahead
	file->f_version = ++event;
#else
	file->f_version++;
#endif
    }
out:
    trace_llseek_post(stoptr(file->f_dentry->d_inode->i_sb), file, offset, origin, err, &context);
    print_exit_status(err);
    return err;
}


/*
 * generic_file_read updates the atime of upper layer inode.
 * But, it doesn't give us a chance to update the atime of
 * the lower layer inode. This function is a wrapper to
 * generic_file_read. it Updates the atime of the lower level
 * inode if generic_file_read returns without any errors
 * This is to be used only for file reads and only if
 * FIST_FILTER_DATA is defined. The function to be
 * used for directory reads is tracefs_read
 */
STATIC ssize_t
tracefs_read_update_atime(file_t *file, char* buf, size_t count, loff_t *ppos)
{
    int err = 0;
    void *context;

    print_entry_location();
    trace_read_pre(stoptr(file->f_dentry->d_inode->i_sb), file, buf, count, ppos, &context);

    err = generic_file_read(file, buf, count, ppos);

    if (err >= 0)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	update_atime(itohi(file->f_dentry->d_inode));
#else
	touch_atime(ftohf(file)->f_vfsmnt,dtohd(file->f_dentry));
#endif

    trace_read_post(stoptr(file->f_dentry->d_inode->i_sb), file, buf, count, ppos, err, &context);

    print_exit_status(err);
    return err;
}

STATIC ssize_t
tracefs_read(file_t *file, char *buf, size_t count, loff_t *ppos)
{
    int err = -EINVAL;
    file_t *hidden_file = NULL;
    void *context;
    loff_t pos = *ppos;

    print_entry_location();
    trace_read_pre(stoptr(file->f_dentry->d_inode->i_sb), file, buf, count, ppos, &context);
    if(ftopd(file) != NULL)
       hidden_file = ftohf(file); /* CPW: Moved to after print_entry_location */
    ASSERT(hidden_file != NULL);

    if (!hidden_file->f_op || !hidden_file->f_op->read)
	goto out;;

    err = hidden_file->f_op->read(hidden_file, buf, count, &pos);
    if (err >= 0) {
	/* atime should also be updated for reads of size zero or more */
	fist_copy_attr_atime(file->f_dentry->d_inode,
			     hidden_file->f_dentry->d_inode);
    }

    // MAJOR HACK
    /*
     * because pread() does not have any way tos tell us that it is
     * our caller, then we don't know for sure if we have to update
     * the file positions.  This hack relies on read() having passed us
     * the "real" pointer of its struct file's f_pos field.
     */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    if (ppos == &file->f_pos)
	hidden_file->f_pos = *ppos = pos;
    if (hidden_file->f_reada) { /* update readahead information if needed */
	file->f_reada = hidden_file->f_reada;
	file->f_ramax = hidden_file->f_ramax;
	file->f_raend = hidden_file->f_raend;
	file->f_ralen = hidden_file->f_ralen;
	file->f_rawin = hidden_file->f_rawin;
    }
    if(ppos == &file->f_pos)
#else
    hidden_file->f_pos = *ppos = pos;
    memcpy(&(file->f_ra),&(hidden_file->f_ra),sizeof(struct file_ra_state));
#endif
out:
    trace_read_post(stoptr(file->f_dentry->d_inode->i_sb), file, buf, count, ppos, err, &context);

    print_exit_status(err);
    return err;
}


STATIC ssize_t
tracefs_write_update_atime(file_t *file, const char *buf, size_t count, loff_t *ppos)
{
    int err = 0;
    void *context;

    print_entry_location();
    trace_write_pre(stoptr(file->f_dentry->d_inode->i_sb), file, buf, count, ppos, &context);

    err = generic_file_write(file, buf, count, ppos);

/*    if (err >= 0)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	update_atime(itohi(file->f_dentry->d_inode));
#else
	touch_atime(ftohf(file)->f_vfsmnt,dtohd(file->f_dentry));
#endif*/

    trace_write_post(stoptr(file->f_dentry->d_inode->i_sb), file, buf, count, ppos, err, &context);
    print_exit_status(err);
    return err;
}


/* this tracefs_write() does not modify data pages! */
STATIC ssize_t
tracefs_write(file_t *file, const char *buf, size_t count, loff_t *ppos)
{
    int err = -EINVAL;
    file_t *hidden_file = NULL;
    inode_t *inode;
    inode_t *hidden_inode;
    void *context;
    loff_t pos = *ppos;

    print_entry_location();
    trace_write_pre(stoptr(file->f_dentry->d_inode->i_sb), file, buf, count, ppos, &context);

    if(ftopd(file) != NULL)
	hidden_file = ftohf(file);
    ASSERT(hidden_file != NULL);

    inode = file->f_dentry->d_inode;
    hidden_inode = itohi(inode);

    if (!hidden_file->f_op || !hidden_file->f_op->write)
	goto out;

    /* adjust for append -- seek to the end of the file */
    if((file->f_flags & O_APPEND) && (count != 0))
	pos = i_size_read(inode);

    if(count)
	err = hidden_file->f_op->write(hidden_file, buf, count, &pos);
    else
	err = 0;

    /*
     * copy ctime and mtime from lower layer attributes
     * atime is unchanged for both layers
     */
    if (err >= 0)
	fist_copy_attr_times(inode, hidden_inode);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    /*
     * XXX: MAJOR HACK
     *
     * because pwrite() does not have any way to tell us that it is
     * our caller, then we don't know for sure if we have to update
     * the file positions.  This hack relies on write() having passed us
     * the "real" pointer of its struct file's f_pos field.
     */
    if (ppos == &file->f_pos)
	hidden_file->f_pos = *ppos = pos;
    /* update this inode's size */
    if (pos > inode->i_size)
	inode->i_size = pos;
    if(ppos == &file->f_pos)
#else
    hidden_file->f_pos = *ppos = pos;
    //update this inode's size
    if(pos > i_size_read(inode))
	i_size_write(inode,pos);
#endif

out:
    trace_write_post(stoptr(file->f_dentry->d_inode->i_sb), file, buf, count, ppos, err, &context);
    print_exit_status(err);
    return err;
}


#if defined(FIST_FILTER_NAME) || defined(FIST_FILTER_SCA)
struct tracefs_getdents_callback {
    void *dirent;
    dentry_t *dentry;
    filldir_t filldir;
    int err;
    int filldir_called;
    int entries_written;
};

/* copied from generic filldir in fs/readir.c */
STATIC int
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,9)
tracefs_filldir(void *dirent, const char *name, int namlen, loff_t offset, ino_t ino, unsigned int d_type)
# else /* kernel older than 2.4.9 */
tracefs_filldir(void *dirent, const char *name, int namlen, off_t offset, ino_t ino, unsigned int d_type)
# endif /* kernel older than 2.4.9 */
{
    struct tracefs_getdents_callback *buf = (struct tracefs_getdents_callback *) dirent;
    int err;
    char err;
    char* decoded_name;
    int decoded_length;

    print_entry_location();
    buf->filldir_called++;

    decoded_length = tracefs_decode_filename(name,namlen,&decoded_name,SKIP_DOTS,buf->dentry->d_inode,buf->dentry->d_sb);

    if(decoded_length < 0)
	return 0;

    err = buf->filldir(buf->dirent, decoded_name, decoded_namlen, offset, ino, d_type);

    KFREE(decoded_name);
    if(err >= 0)
	buf->entries_written++;

    print_exit_status(err);
    return err;
}
#endif /* FIST_FILTER_NAME || FIST_FILTER_SCA */


#define COUNT_OFFSET(x) ((char *)(x) + 2 * sizeof(void *))
STATIC int
tracefs_readdir(file_t *file, void *dirent, filldir_t filldir)
{
    int err = -ENOTDIR;
    file_t *hidden_file = NULL;
    inode_t *inode;
    void *context;
    int count = *(int *)COUNT_OFFSET(dirent);
#if defined(FIST_FILTER_NAME) || defined(FIST_FILTER_SCA)
    struct tracefs_getdents_callback buf;
#endif /* FIST_FILTER_NAME || FIST_FILTER_SCA */

    print_entry_location();
    trace_readdir_pre(stoptr(file->f_dentry->d_inode->i_sb), file, dirent, filldir, count, &context);
    if(ftopd(file) != NULL)
	hidden_file = ftohf(file); /* CPW: Moved to after print_entry_location */
    ASSERT(hidden_file != NULL);

    inode = file->f_dentry->d_inode;

    fist_checkinode(inode, "tracefs_readdir");
    hidden_file->f_pos = file->f_pos;

#if defined(FIST_FILTER_NAME) || defined(FIST_FILTER_SCA)
    /* prepare for callback */
    buf.dirent = dirent;
    buf.dentry = file->f_dentry;
    buf.filldir = filldir;
 retry:
    buf.filldir_called = 0;
    buf.entries_written = 0;
    buf.err = 0;
    err = vfs_readdir(hidden_file, tracefs_filldir, (void *) &buf);
    if(buf.err)
	err = buf.err;
    if(buf.filldir_called && !buf.entries_written)
	goto retry;
#else /* not FIST_FILTER_NAME */
    err = vfs_readdir(hidden_file, filldir, dirent);
#endif /* not FIST_FILTER_NAME */

    file->f_pos = hidden_file->f_pos;
    if (err > 0)
	fist_copy_attr_atime(inode, hidden_file->f_dentry->d_inode);

    fist_checkinode(inode, "post tracefs_readdir");
    count = *(int *)COUNT_OFFSET(dirent);
    trace_readdir_post(stoptr(file->f_dentry->d_inode->i_sb), file, dirent, filldir, count, err, &context);
    print_exit_status(err);
    return err;
}


STATIC unsigned int
tracefs_poll(file_t *file, poll_table *wait)
{
    unsigned int mask = DEFAULT_POLLMASK;
    file_t *hidden_file = NULL;
    void *context;

    print_entry_location();
    trace_poll_pre(stoptr(file->f_dentry->d_inode->i_sb), file, wait, &context);
    if(ftopd(file) != NULL)
	hidden_file = ftohf(file); /* CPW: Moved to after print_entry_location */
    ASSERT(hidden_file != NULL);

    if (!hidden_file->f_op || !hidden_file->f_op->poll)
	goto out;

    mask = hidden_file->f_op->poll(hidden_file, wait);

out:
    trace_poll_post(stoptr(file->f_dentry->d_inode->i_sb), file, wait, mask, &context);
    print_exit_status(mask);
    return mask;
}

STATIC int
tracefs_ioctl(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg)
{
    int err = 0;		/* don't fail by default */
    file_t *hidden_file = NULL;
    vfs_t *this_vfs;
    vnode_t *this_vnode;
    void *context;
    int val;
#ifdef FIST_COUNT_WRITES
    extern unsigned long count_writes, count_writes_middle;
# endif /* FIST_COUNT_WRITES */

    print_entry_location();

    this_vfs = inode->i_sb;
    this_vnode = inode;

    /* check if asked for local commands */
    switch (cmd) {
    case FIST_IOCTL_GET_DEBUG_VALUE:
	val = fist_get_debug_value();
	printk("IOCTL GET: send arg %d\n", val);
	err = put_user(val, (int *) arg);
#ifdef FIST_COUNT_WRITES
	printk("COUNT_WRITES:%lu:%lu\n", count_writes, count_writes_middle);
# endif /* FIST_COUNT_WRITES */
	break;

    case FIST_IOCTL_SET_DEBUG_VALUE:
	err = get_user(val, (int *) arg);
	if (err)
	    break;
	fist_dprint(6, "IOCTL SET: got arg %d\n", val);
	if (val < 0 || val > 20) {
	    err = -EINVAL;
	    break;
	}
	fist_set_debug_value(val);
	break;

    case TRACEFS_IOCTL_ADD_OUTPUT_DRIVERS:
	err = tracefs_ioctl_add_odrivers(inode, file, cmd, arg);
	break;
    case TRACEFS_IOCTL_ADD_ASSEMBLY_DRIVERS:
	err = tracefs_ioctl_add_adrivers(inode, file, cmd, arg);
	break;
    case TRACEFS_IOCTL_ADD_RULE:
	err = tracefs_ioctl_add_rule(inode, file, cmd, arg);
	break;
    case TRACEFS_IOCTL_DEL_OUTPUT_DRIVERS:
	err = tracefs_ioctl_del_odrivers(inode, file, cmd, arg);
	break;
    case TRACEFS_IOCTL_DEL_ASSEMBLY_DRIVERS:
	err = tracefs_ioctl_del_adrivers(inode, file, cmd, arg);
	break;
    case TRACEFS_IOCTL_DEL_RULE:
	err = tracefs_ioctl_del_rule(inode, file, cmd, arg);
	break;
    case TRACEFS_IOCTL_TEARDOWN:
	err = tracefs_ioctl_teardown(inode, file, cmd, arg);
	break;

    case TRACEFS_IOCTL_START_TRACE:
	err  = tracefs_ioctl_start_trace(inode, file, cmd, arg);
	break;

    case TRACEFS_IOCTL_STOP_TRACE:
	err = tracefs_ioctl_stop_trace(inode, file, cmd, arg);
	break;

    default:
	/* only trace the non-stacked ioctl */
	trace_ioctl_pre(stoptr(inode->i_sb), inode, file, cmd, arg, &context);
	hidden_file = ftohf(file);
	/* pass operation to hidden filesystem, and return status */
	if (hidden_file->f_op && hidden_file->f_op->ioctl)
	    err = hidden_file->f_op->ioctl(itohi(inode), hidden_file, cmd, arg);
	else
	    err = -ENOTTY;	/* this is an unknown ioctl */
	trace_ioctl_post(stoptr(inode->i_sb), inode, file, cmd, arg, err, &context);
    } /* end of outer switch statement */

    print_exit_status(err);
    return err;
}


/* FIST-LITE special version of mmap */
STATIC int
tracefs_mmap(file_t *file, vm_area_t *vma)
{
    int err = 0;
    file_t *hidden_file;
//    inode_t *inode;
//    inode_t *lower_inode;
    void *context;

    print_entry_location();
    trace_mmap_pre(stoptr(file->f_dentry->d_inode->i_sb), file, vma, &context);

    hidden_file = ftohf(file); /* CPW: Moved to after print_entry_location */

    ASSERT(hidden_file != NULL);
    if(!hidden_file->f_op || !hidden_file->f_op->mmap) {
	err = -ENODEV;
	goto out;
    }
    ASSERT(hidden_file->f_op != NULL);
    ASSERT(hidden_file->f_op->mmap != NULL);

    vma->vm_file = hidden_file;
    err = hidden_file->f_op->mmap(hidden_file, vma);
    get_file(hidden_file); /* make sure it doesn't get freed on us */

 out:
    trace_mmap_post(stoptr(file->f_dentry->d_inode->i_sb), file, vma, err, &context);
    fput(file);			/* no need to keep extra ref on ours */

    print_exit_status(err);
    return err;
}


STATIC int
tracefs_open(inode_t *inode, file_t *file)
{
    int err = 0;
    int hidden_flags;
    file_t *hidden_file;
    void *context;
    dentry_t *hidden_dentry = NULL;

    print_entry_location();

    trace_open_pre(stoptr(inode->i_sb), inode, file, &context);
    ftopdlh(file) = KMALLOC(sizeof(struct tracefs_file_info *), GFP_KERNEL);
    if (!ftopd(file)) {
	err = -ENOMEM;
	goto out;
    }

    hidden_dentry = tracefs_hidden_dentry(file->f_dentry); /* CPW: Moved after print_entry */
    fist_print_dentry("tracefs_open IN hidden_dentry", hidden_dentry);
    dget(hidden_dentry);
    hidden_flags = file->f_flags;
    if (mmap_support(inode->i_sb)) {
	/* If a file is opened for write only, make read-write. */
	if ((hidden_flags & O_ACCMODE) == O_WRONLY)
	    hidden_flags = (hidden_flags & ~O_ACCMODE) | O_RDWR;
	if (file->f_flags & O_APPEND) {
	    fist_dprint(5, "file is opened in append-only mode\n");
	    hidden_flags &= ~O_APPEND; /* turn off O_APPEND flag */
	}
    }

    /*
     * dentry_open will decrement mnt refcnt if err.
     * otherwise fput() will do an mntput() for us upon file close.
     */
    mntget(stopd(inode->i_sb)->hidden_mnt);
    hidden_file = dentry_open(hidden_dentry, stopd(inode->i_sb)->hidden_mnt, hidden_flags);
    if (IS_ERR(hidden_file)) {
	err = PTR_ERR(hidden_file);
	goto out;
    }
    ftohf(file) = hidden_file;	/* link two files */

 out:
    if (err < 0 && ftopd(file)) {
	KFREE(ftopd(file));
    }
    trace_open_post(stoptr(inode->i_sb), inode, file, err, &context);
    fist_print_dentry("tracefs_open OUT hidden_dentry", hidden_dentry);
    print_exit_status(err);
    return err;
}


STATIC int
tracefs_flush(file_t *file)
{
    int err = 0;		/* assume ok (see open.c:close_fp) */
    file_t *hidden_file = NULL;
    void *context;

    print_entry_location();
    trace_flush_pre(stoptr(file->f_dentry->d_inode->i_sb), file, &context);

    if(ftopd(file) != NULL)
	hidden_file = ftohf(file); /* CPW: Moved after print_entry_location */
    ASSERT(hidden_file != NULL);

    if (!hidden_file->f_op || !hidden_file->f_op->flush)
	goto out;

    err = hidden_file->f_op->flush(hidden_file);

out:
    trace_flush_post(stoptr(file->f_dentry->d_inode->i_sb), file, err, &context);
    print_exit_status(err);
    return err;
}


STATIC int
tracefs_release(inode_t *inode, file_t *file)
{
    int err = 0;
    file_t *hidden_file = NULL;
    dentry_t *hidden_dentry;
    inode_t *hidden_inode = NULL;
    void *context;

    print_entry_location();
    trace_release_pre(stoptr(inode->i_sb), inode, file, &context);

    if(ftopd(file) != NULL)
	hidden_file = ftohf(file);
    KFREE(ftopd(file));
    ASSERT(hidden_file != NULL);

    ASSERT(inode);
    hidden_inode = itohi(inode);
    ASSERT(hidden_inode);

    fist_print_dentry("tracefs_release IN hidden_dentry", hidden_file->f_dentry);
    fist_checkinode(inode, "tracefs_release");
    fist_dprint(6, "tracefs_release IN, file->f_count=%d\n", file->f_count);
    /*
     * will decrement file refcount, and if 0, destroy the file,
     * which will call the lower file system's file release function.
     */
    hidden_dentry = hidden_file->f_dentry;
    fput(hidden_file);

    fist_copy_attr_all(inode, hidden_inode); //XXX: Tim: is this needed - not in templates, but might be necessary for accurate traces

    fist_dprint(6, "tracefs_release done\n");
    fist_checkinode(inode, "post tracefs_release");

    fist_print_dentry("tracefs_release OUT hidden_dentry", hidden_dentry);
    trace_release_post(stoptr(inode->i_sb), inode, file, err, &context);
    print_exit_status(err);
    return err;
}


STATIC int
tracefs_fsync(file_t *file, dentry_t *dentry, int datasync)
{
    int err = -EINVAL;
    file_t *hidden_file = NULL;
    dentry_t *hidden_dentry;
    void *context;

    print_entry_location();
    trace_fsync_pre(stoptr(file->f_dentry->d_inode->i_sb), file, dentry, datasync, &context);

    if(file == NULL) {
	hidden_dentry = tracefs_hidden_dentry(dentry);
	if(hidden_dentry->d_inode->i_fop && hidden_dentry->d_inode->i_fop->fsync) {

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	    down(&hidden_dentry->d_inode->i_sem);
#else
      mutex_lock(&hidden_dentry->d_inode->i_mutex);
#endif

	    err = hidden_dentry->d_inode->i_fop->fsync(hidden_file,hidden_dentry,datasync);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	    up(&hidden_dentry->d_inode->i_sem);
#else
      mutex_unlock(&hidden_dentry->d_inode->i_mutex);
#endif


	}
    }
    else {
	if(ftopd(file) != NULL) {
	    hidden_file = ftohf(file);
	    ASSERT(hidden_file != NULL);

	    hidden_dentry = tracefs_hidden_dentry(dentry);

	    if(hidden_file->f_op && hidden_file->f_op->fsync) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
  		down(&hidden_dentry->d_inode->i_sem);
#else
      mutex_lock(&hidden_dentry->d_inode->i_mutex);
#endif
		err = hidden_file->f_op->fsync(hidden_file,hidden_dentry,datasync);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
		up(&hidden_dentry->d_inode->i_sem);
#else
      mutex_unlock(&hidden_dentry->d_inode->i_mutex);
#endif
	    }
	}
    }

    trace_fsync_post(stoptr(file->f_dentry->d_inode->i_sb), file, dentry, datasync, err, &context);
    print_exit_status(err);
    return err;
}


STATIC int
tracefs_fasync(int fd, file_t *file, int flag)
{
    int err = 0;
    file_t *hidden_file = NULL;
    void *context;

    print_entry_location();
    trace_fasync_pre(stoptr(file->f_dentry->d_inode->i_sb), fd, file, flag, &context);
    if(ftopd(file) != NULL)
	hidden_file = ftohf(file); /* CPW: Moved to after print_entry_location */
    ASSERT(hidden_file != NULL);

    if (hidden_file->f_op && hidden_file->f_op->fasync)
	err = hidden_file->f_op->fasync(fd, hidden_file, flag);

    trace_fasync_post(stoptr(file->f_dentry->d_inode->i_sb), fd, file, flag, err, &context);
    print_exit_status(err);
    return err;
}

static inline void __locks_delete_block(struct file_lock *waiter)
{
        list_del_init(&waiter->fl_block);
        list_del_init(&waiter->fl_link);
        waiter->fl_next = NULL;
}

static void locks_delete_block(struct file_lock *waiter)
{
        lock_kernel();
        __locks_delete_block(waiter);
        unlock_kernel();
}

STATIC inline int
tracefs_posix_lock(file_t *file, struct file_lock *fl, int cmd)
{
    int error;
    for(;;) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	error = posix_lock_file(file, fl, cmd);
#else
        error = posix_lock_file(file, fl);
#endif
        if ((error != -EAGAIN) || (cmd == F_SETLK))
            break;

        error = wait_event_interruptible(fl->fl_wait, !fl->fl_next);
        if(!error)
            continue;

        locks_delete_block(fl);
        break;
    }

    return error;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
STATIC int
tracefs_setlk(file_t *file, int cmd, struct file_lock *fl)
{
    int error;
    inode_t *inode,*hidden_inode;
    file_t *hidden_file = NULL;

    print_entry_location();

    error = -EINVAL;
    if (ftopd(file) != NULL)   // XXX: ZH: this is wrong?: in this case ASSERT is not in ifndef... Put ASSERT before the if statement, kill if statement
	hidden_file = ftohf(file);
    ASSERT(hidden_file != NULL);

    inode = file->f_dentry->d_inode;
    hidden_inode = hidden_file->f_dentry->d_inode;

    /* Don't allow mandatory locks on files that may be memory mapped
     * and shared.
     */
    if (IS_MANDLOCK(hidden_inode) &&
       (hidden_inode->i_mode & (S_ISGID | S_IXGRP)) == S_ISGID &&
       mapping_writably_mapped(hidden_file->f_mapping)) {
                error = -EAGAIN;
                goto out;
    }

    if (cmd == F_SETLKW) {
    	fl->fl_flags |= FL_SLEEP;
    }

    error = -EBADF;
    switch (fl->fl_type) {
        case F_RDLCK:
            if (!(hidden_file->f_mode & FMODE_READ))
                goto out;
            break;
        case F_WRLCK:
            if (!(hidden_file->f_mode & FMODE_WRITE))
                goto out;
            break;
        case F_UNLCK:
            break;
        default:
            error = -EINVAL;
            goto out;
        }

        fl->fl_file = hidden_file;
        error = security_file_lock(hidden_file, fl->fl_type);
        if (error)
                goto out;

        if (hidden_file->f_op && hidden_file->f_op->lock != NULL) {
            error = hidden_file->f_op->lock(hidden_file, cmd, fl);
	    if(error)
		goto out;
	    goto upper_lock;
        }

	error = tracefs_posix_lock(file,fl,cmd);
	if(error)
	    goto out;

upper_lock:
    fl->fl_file = file;
    error = tracefs_posix_lock(file, fl, cmd);
    if (error) {
        fl->fl_type = F_UNLCK;
        fl->fl_file = hidden_file;
        tracefs_posix_lock(hidden_file, fl, cmd);
    }

out:
	print_exit_status(error);
        return error;
}


STATIC int
tracefs_getlk(file_t *file, struct file_lock *fl)
{
    int error=0;
    struct file_lock *tempfl=NULL;

    print_entry_location();

    if (file->f_op && file->f_op->lock) {
        error = file->f_op->lock(file, F_GETLK, fl);
        if (error < 0)
            goto out;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11)
        else if (error == LOCK_USE_CLNT)
            /* Bypass for NFS with no locking - 2.0.36 compat */
            tempfl = posix_test_lock(file, fl);
#endif
    } else {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
        tempfl = posix_test_lock(file, fl);
#else
        error = posix_test_lock(file, fl,tempfl);
#endif
    }

    if (!tempfl)
        fl->fl_type = F_UNLCK;
    else
        memcpy(fl, tempfl, sizeof(struct file_lock));

out:
    print_exit_status(error);
    return error;
}
#endif

STATIC int
tracefs_lock(file_t *file, int cmd, struct file_lock *fl)
{
    int err = 0;
    file_t *hidden_file = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    struct file_lock *tmpfl = NULL;
#endif
    void *context;

    print_entry_location();
    trace_lock_pre(stoptr(file->f_dentry->d_inode->i_sb), file, cmd, fl, &context);

    if(ftopd(file) != NULL)
	hidden_file = ftohf(file); /* CPW: Moved to after print_entry_location */
    ASSERT(hidden_file != NULL);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    if (hidden_file->f_op->lock) {
        err = hidden_file->f_op->lock(hidden_file, F_GETLK, fl);
    } else {
        posix_test_lock(hidden_file, fl);
    }
#else
    print_exit_status(err);
    return err;

    err = -EINVAL;
    if(!fl)
	goto out;

    fl->fl_file = hidden_file;
    switch(cmd) {
    case F_GETLK:
    case F_GETLK64:
	err = tracefs_getlk(hidden_file, fl);
	break;

    case F_SETLK:
    case F_SETLKW:
    case F_SETLK64:
    case F_SETLKW64:
	fl->fl_file = file;
	err = tracefs_setlk(hidden_file, cmd, fl);
	break;

    default:
	err = -EINVAL;
    }

    fl->fl_file = file;

 out:
#endif
    trace_lock_post(stoptr(file->f_dentry->d_inode->i_sb), file, cmd, fl, err, &context);
    print_exit_status(err);
    return err;
}



struct file_operations tracefs_dir_fops =
{
    llseek:	tracefs_llseek,
    read:	tracefs_read,
    write:	tracefs_write,
    readdir:	tracefs_readdir,
    poll:	tracefs_poll,
    ioctl:	tracefs_ioctl,
    mmap:	tracefs_mmap,
    open:	tracefs_open,
    flush:	tracefs_flush,
    release:	tracefs_release,
    fsync:	tracefs_fsync,
    fasync:	tracefs_fasync,
    lock:	tracefs_lock,
    /* not needed: readv */
    /* not needed: writev */
    /* not implemented: sendpage */
    /* not implemented: get_unmapped_area */
};

struct file_operations tracefs_main_fops =
{
    llseek:	tracefs_llseek,
    read:	tracefs_read,
    write:	tracefs_write,
    readdir:	tracefs_readdir,
    poll:	tracefs_poll,
    ioctl:	tracefs_ioctl,
    mmap:	tracefs_mmap,
    open:	tracefs_open,
    flush:	tracefs_flush,
    release:	tracefs_release,
    fsync:	tracefs_fsync,
    fasync:	tracefs_fasync,
    lock:	tracefs_lock,
    /* not needed: readv */
    /* not needed: writev */
    /* not implemented: sendpage */
    /* not implemented: get_unmapped_area */
};

struct file_operations tracefs_mmap_fops =
{
    llseek:	tracefs_llseek,
    read:	tracefs_read_update_atime,
    write:	tracefs_write_update_atime,
    readdir:	tracefs_readdir,
    poll:	tracefs_poll,
    ioctl:	tracefs_ioctl,
    mmap:	generic_file_mmap,
    open:	tracefs_open,
    flush:	tracefs_flush,
    release:	tracefs_release,
    fsync:	tracefs_fsync,
    fasync:	tracefs_fasync,
    lock:	tracefs_lock,
    /* not needed: readv */
    /* not needed: writev */
    /* not implemented: sendpage */
    /* not implemented: get_unmapped_area */
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 * vim:shiftwidth=4
 */
