/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: inode.c,v 1.32 2007/01/30 06:30:53 gopalas Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#ifdef FISTGEN
# include "fist_tracefs.h"
#endif /* FISTGEN */
#include "fist.h"
#include "tracefs.h"
#include "tracer.h"

extern int tracefs_fill_zeros(file_t*,page_t*,unsigned);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/namei.h>
#include <linux/fs.h>


/*
 * Whee.. Deadlock country. Happily there are only two VFS
 * operations that does this..
 */
static inline void double_down(struct semaphore *s1, struct semaphore *s2)
{
        if (s1 != s2) {
                if ((unsigned long) s1 < (unsigned long) s2) {
                        struct semaphore *tmp = s2;
                        s2 = s1; s1 = tmp;
                }
                down(s1);
        }
        down(s2);
}

static inline void double_up(struct semaphore *s1, struct semaphore *s2)
{
        up(s1);
        if (s1 != s2)
                up(s2);
}
/*Adding new functions double_mutex_lock and double_mutex_unlock for inode mutexes*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
static inline void double_mutex_lock(struct mutex *m1, struct mutex *m2)
{
        if (m1 != m2)
        {
          if ((unsigned long) m1 < (unsigned long) m2)
          {
            struct mutex *tmp = m2;
             m2 = m1; m1 = tmp;
          }
          mutex_lock(m1);
        }
        mutex_lock(m2);
}

static inline void double_mutex_unlock(struct mutex *m1, struct mutex *m2)
{
        mutex_unlock(m1);
        if (m1 != m2)
                mutex_unlock(m2);
}
#endif
static inline void triple_up(struct semaphore *s1,
                             struct semaphore *s2,
                             struct semaphore *s3)
{
        up(s1);
        if (s1 != s2)
                up(s2);
        up(s3);
}

static inline void double_lock(struct dentry *d1, struct dentry *d2)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
    double_down(&d1->d_inode->i_sem, &d2->d_inode->i_sem);
#else
    double_mutex_lock(&d1->d_inode->i_mutex, &d2->d_inode->i_mutex);
#endif
}

static inline void double_unlock(struct dentry *d1, struct dentry *d2)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
    double_up(&d1->d_inode->i_sem,&d2->d_inode->i_sem);
#else
    double_mutex_unlock(&d1->d_inode->i_mutex, &d2->d_inode->i_mutex);
#endif
        dput(d1);
        dput(d2);
}

static inline struct dentry *lock_parent(struct dentry *dentry)
{
    struct dentry *dir = dget(dentry->d_parent);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
    down(&dir->d_inode->i_sem);
#else
    mutex_lock(&dir->d_inode->i_mutex);
#endif
    return dir;
}

static inline struct dentry *get_parent(struct dentry *dentry) {
        return dget(dentry->d_parent);
}

static inline void unlock_dir(struct dentry *dir)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
    up(&dir->d_inode->i_sem);
#else
    mutex_unlock(&dir->d_inode->i_mutex);
#endif
    dput(dir);
}

#endif
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif
/*
 * Function to handle truncations extending the size of the file
 */
STATIC int
tracefs_truncate(dentry_t *dentry, loff_t new_length)
{
        inode_t *inode = dentry->d_inode;
        dentry_t *lower_dentry;
        file_t fake_file, *lower_file = NULL;
        page_t *tmp_page;
        int index;
        unsigned to;
        int err = 0;

        print_entry_location();

        memset(&fake_file, 0, sizeof(fake_file));

        /*
         * if this is a truncate past the end of the file, call tracefs_fill_zeros
         */
        if (new_length > i_size_read(inode)) {
                fake_file.f_dentry = dentry;
                /* allocate space for private data */
                ftopdlh(&fake_file) = KMALLOC(sizeof(struct tracefs_file_info), GFP_KERNEL);
                if (!ftopd(&fake_file)) {
                        err = -ENOMEM;
                        goto out;
                }
                /*
                 * dentry_open will decrement mnt refcnt if err.
                 * otherwise fput() will do an mntput() for us upon file close.
                 */
                lower_dentry = tracefs_hidden_dentry(dentry);
		dget(lower_dentry);
                mntget(stopd(inode->i_sb)->hidden_mnt);
                lower_file = dentry_open(lower_dentry, stopd(inode->i_sb)->hidden_mnt, O_RDWR);
                if (IS_ERR(lower_file)) {
                        err = PTR_ERR(lower_file);
                        goto out_free;
                }
                ftohf(&fake_file) = lower_file;	/* link two files */

                index = new_length >> PAGE_CACHE_SHIFT;

                tmp_page = tracefs_get1page(&fake_file, index);
                if (IS_ERR(tmp_page)) {
                        err = PTR_ERR(tmp_page);
                        goto out_fput;
                }
                kmap(tmp_page);

                to = new_length & ~PAGE_CACHE_MASK;

                /*
                 * 'to' is passed to tracefs_fill_zeros instead of a 'from'
                 *  because unlike the normal case where we fill in zeros
                 *  till 'from', we need to fill it in with zeros till the
                 *  end of the file
                 */
                err = tracefs_fill_zeros(&fake_file, tmp_page, to);
                kunmap(tmp_page);
        }

out_fput:
        if (lower_file)
                filp_close(lower_file, NULL);
out_free:
        if (ftopd(&fake_file))
                KFREE(ftopd(&fake_file));
out:
        print_exit_status(err);
        return err;
}

STATIC int
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
tracefs_create(inode_t *dir, dentry_t *dentry, int mode)
#else
tracefs_create(inode_t *dir, dentry_t *dentry, int mode,struct nameidata* nd)
#endif
{
    int err;
    dentry_t *hidden_dentry;
    void *context;

    dentry_t *hidden_dir_dentry;

    print_entry_location();
    trace_create_pre(stoptr(dir->i_sb), dir, dentry, mode, &context);
    hidden_dentry = tracefs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    ASSERT(hidden_dentry != NULL);
    fist_checkinode(dir, "tracefs_create");

    hidden_dir_dentry = lock_parent(hidden_dentry);

    err = PTR_ERR(hidden_dir_dentry);
    if (IS_ERR(hidden_dir_dentry))
	goto out;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    err = vfs_create(hidden_dir_dentry->d_inode, hidden_dentry, mode);
#else
    err = vfs_create(hidden_dir_dentry->d_inode, hidden_dentry, mode,nd);
#endif
    /* XXX this could potentially return a negative hidden_dentry! */
    if (err)
	goto out_lock;

    err = tracefs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out_lock;

    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);


 out_lock:
    unlock_dir(hidden_dir_dentry);
 out:
    fist_checkinode(dir, "post tracefs_create");
    trace_create_post(stoptr(dir->i_sb), dir, dentry, mode, err, &context);
    print_exit_status(err);
    return err;
}

STATIC dentry_t *
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
tracefs_lookup(inode_t *dir, dentry_t *dentry)
#else
tracefs_lookup(inode_t* dir,dentry_t *dentry,struct nameidata* nd)
#endif
{
    int err = 0;
    dentry_t *hidden_dir_dentry;
    dentry_t *hidden_dentry = NULL;
    void *context;
    const char *name;
    vnode_t *this_vnode;
    dentry_t *this_dir;
    unsigned int namelen;
    char* encoded_name;
    int encoded_namelen;

    print_entry_location();
    trace_lookup_pre(stoptr(dir->i_sb), dir, dentry, &context);
    hidden_dir_dentry = tracefs_hidden_dentry(dentry->d_parent); /* CPW: Moved below print_entry_location */
    name = dentry->d_name.name;
    namelen = dentry->d_name.len;
    fist_checkinode(dir, "tracefs_lookup");

    this_vnode = dir;
    this_dir = hidden_dir_dentry;

    fist_print_dentry("tracefs_lookup IN", dentry);
    fist_print_dentry("tracefs_lookup: dentry->d_parent IN", dentry->d_parent);
    fist_print_dentry("tracefs_lookup: hidden_dir_dentry IN", hidden_dir_dentry);
    fist_print_inode("tracefs_lookup: dir IN", dir);

    if (hidden_dir_dentry->d_inode)
	fist_print_inode("tracefs_lookup: hidden_dir_dentry->d_inode",
			 hidden_dir_dentry->d_inode);

    /* must initialize dentry operations */
    dentry->d_op = &tracefs_dops;

    ;

    /* increase refcount of base dentry (lookup_one[_len] will decrement) */
    // THIS IS RIGHT! (don't "fix" it)
    // NO THIS IS WRONG IN 2.3.99-pre6. lookup_one[_len] will NOT decrement
    // dget(hidden_dir_dentry);

    encoded_namelen = tracefs_encode_filename(name,
					      namelen,
					      &encoded_name,
					      SKIP_DOTS,dir,dir->i_sb);

    /*if there was an error in encoding filenames*/
    if(encoded_namelen < 0) {
	err = encoded_namelen;
	goto out_free;
    }

    /* will allocate a new hidden dentry if needed */
    hidden_dentry = lookup_one_len(encoded_name, hidden_dir_dentry, encoded_namelen - 1);

    if (IS_ERR(hidden_dentry)) {
	/*
	 * this produces an unusual dentry: one that has neither an
	 * inode, nor a private structure attached to it. All cleanup
	 * methods (d_delete, d_release, etc) must be prepared to deal
	 * with such dentries. Ion 09/29/2001
	 */
	printk("ERR from hidden_dentry!!!\n");
	err = PTR_ERR(hidden_dentry);
	goto out;
    }

    ;

    /* update parent directory's atime */
    fist_copy_attr_atime(dir, hidden_dir_dentry->d_inode);
    /* link the upper and lower dentries */
    dtopdlh(dentry) = (struct tracefs_dentry_info *) KMALLOC(sizeof(struct tracefs_dentry_info), GFP_KERNEL);
    if (!dtopd(dentry)) {
	err = -ENOMEM;
	goto out_dput;
    }
    ((struct tracefs_dentry_info *) dtopdlh(dentry))->wdi_dentry = hidden_dentry;


    /* lookup is special: it needs to handle negative dentries */
    if (!hidden_dentry->d_inode) {
	d_add(dentry, NULL);
	fist_print_dentry("tracefs_lookup OUT hidden_dentry", hidden_dentry);
	goto out;
    }

    fist_dprint(6, "lookup \"%s\" -> inode %d\n", name, hidden_dentry->d_inode->i_ino);

    err = tracefs_interpose(hidden_dentry, dentry, dir->i_sb, 1);
    if (err)
	goto out_free;

    fist_checkinode(dentry->d_inode, "tracefs_lookup OUT: dentry->d_inode:");
    fist_checkinode(dir, "tracefs_lookup OUT: dir");
    fist_print_dentry("tracefs_lookup OUT hidden_dentry", hidden_dentry);
    fist_print_inode("tracefs_lookup OUT hidden_inode", hidden_dentry->d_inode);


    /* All is well */
    goto out;

 out_free:
    d_drop(dentry);		/* so that our bad dentry will get destroyed */
    KFREE(dtopd(dentry));
    dtopdlh(dentry) = NULL;	/* be safe */

 out_dput:
    dput(hidden_dentry);

 out:
    fist_print_dentry("tracefs_lookup OUT", dentry);
    trace_lookup_post(stoptr(dir->i_sb), dir, dentry, ERR_PTR(err), &context);
    print_exit_status(err);
    return ERR_PTR(err);
}


STATIC int
tracefs_link(dentry_t *old_dentry, inode_t *dir, dentry_t *new_dentry)
{
    int err;
    dentry_t *hidden_old_dentry;
    dentry_t *hidden_new_dentry;
    dentry_t *hidden_dir_dentry;
    void *context;

    print_entry_location();
    trace_link_pre(stoptr(dir->i_sb), old_dentry, dir, new_dentry, &context);
    hidden_old_dentry = tracefs_hidden_dentry(old_dentry); /* CPW: Moved below print_entry_location */
    hidden_new_dentry = tracefs_hidden_dentry(new_dentry);

    fist_checkinode(dir, "tracefs_link-dir");
    fist_checkinode(old_dentry->d_inode, "tracefs_link-oldinode");

    dget(hidden_old_dentry);
    dget(hidden_new_dentry);
    hidden_dir_dentry = lock_parent(hidden_new_dentry);

    err = vfs_link(hidden_old_dentry,
		   hidden_dir_dentry->d_inode,
		   hidden_new_dentry);
    if (err || !hidden_new_dentry->d_inode)
	goto out_lock;

    err = tracefs_interpose(hidden_new_dentry, new_dentry, dir->i_sb, 0);
    if (err)
	goto out_lock;

    fist_copy_attr_timesizes(dir, hidden_new_dentry->d_inode);
    /* propagate number of hard-links */
    old_dentry->d_inode->i_nlink = itohi(old_dentry->d_inode)->i_nlink;

 out_lock:
    unlock_dir(hidden_dir_dentry);
    dput(hidden_new_dentry);
    dput(hidden_old_dentry);
    if (!new_dentry->d_inode)
	d_drop(new_dentry);

    trace_link_post(stoptr(dir->i_sb), old_dentry, dir, new_dentry, err, &context);
    print_exit_status(err);
    return err;
}


STATIC int
tracefs_unlink(inode_t *dir, dentry_t *dentry)
{
    int err = 0;
    inode_t *hidden_dir;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;
    void *context;

    print_entry_location();
    trace_unlink_pre(stoptr(dir->i_sb), dir, dentry, &context);

    hidden_dir = itohi(dir); /* CPW: Moved below print_entry_location */
    hidden_dentry = tracefs_hidden_dentry(dentry);

    ASSERT(hidden_dentry != NULL);
    fist_checkinode(dir, "tracefs_unlink-dir");

    dget(dentry);
    hidden_dir_dentry = lock_parent(hidden_dentry);

    /* avoid destroying the hidden inode if the file is in use */
    dget(hidden_dentry);
    err = vfs_unlink(hidden_dir, hidden_dentry);
    dput(hidden_dentry);

    if (!err)			/* vfs_unlink does that */
	d_delete(hidden_dentry);

    //out_lock:
    fist_copy_attr_times(dir, hidden_dir);
    /* propagate number of hard-links */
    dentry->d_inode->i_nlink = itohi(dentry->d_inode)->i_nlink;
    fist_copy_attr_ctime(dentry->d_inode,dir);

    unlock_dir(hidden_dir_dentry);

    /*
     * call d_drop so the system "forgets" about us
     */
    if (!err) {
	d_drop(dentry);
    }

    dput(dentry);

    fist_checkinode(dir, "post tracefs_unlink-dir");
    trace_unlink_post(stoptr(dir->i_sb), dir, dentry, err, &context);
    print_exit_status(err);
    return err;
}


STATIC int
tracefs_symlink(inode_t *dir, dentry_t *dentry, const char *symname)
{
    int err;
    dentry_t *hidden_dentry = NULL;
    dentry_t *hidden_dir_dentry = NULL;
    umode_t mode;
    char *encoded_symname;
    char encoded_symlen;
    void *context;

    print_entry_location();
    encoded_symlen = tracefs_encode_filename(symname,
					     strlen(symname),
					     &encoded_symname,
					     DO_DOTS,dir,dir->i_sb);

    /* if length is negative, return back */
    if(encoded_symlen < 0) {
	err = encoded_symlen;
	goto out_lock;
    }

    trace_symlink_pre(stoptr(dir->i_sb), dir, dentry, encoded_symname, &context);
    hidden_dentry = tracefs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    fist_checkinode(dir, "tracefs_symlink-dir");

    dget(hidden_dentry);
    hidden_dir_dentry = lock_parent(hidden_dentry);

    mode = S_IALLUGO;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    err = vfs_symlink(hidden_dir_dentry->d_inode,
		      hidden_dentry,
		      encoded_symname);
#else
    err = vfs_symlink(hidden_dir_dentry->d_inode,
		      hidden_dentry,
		      encoded_symname,
		      mode);
#endif

    if (err || !hidden_dentry->d_inode)
	goto out_lock;
    err = tracefs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out_lock;

    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);
    fist_checkinode(dir, "post tracefs_symlink-dir");

 out_lock:
    unlock_dir(hidden_dir_dentry);
    dput(hidden_dentry);
    if (!dentry->d_inode)
	d_drop(dentry);

    trace_symlink_post(stoptr(dir->i_sb), dir, dentry, encoded_symname, err, &context);
    KFREE(encoded_symname);
    print_exit_status(err);
    return err;
}


STATIC int
tracefs_mkdir(inode_t *dir, dentry_t *dentry, int mode)
{
    int err;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;
    void *context;

    print_entry_location();
    trace_mkdir_pre(stoptr(dir->i_sb), dir, dentry, mode, &context);
    hidden_dentry = tracefs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    fist_checkinode(dir, "tracefs_mkdir-dir");

    hidden_dir_dentry = lock_parent(hidden_dentry);

    err = vfs_mkdir(hidden_dir_dentry->d_inode,
		    hidden_dentry,
		    mode);
    if (err || !hidden_dentry->d_inode)
	goto out;

    err = tracefs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out;

    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);
    /* update number of links on parent directory */
    dir->i_nlink = hidden_dir_dentry->d_inode->i_nlink;

    fist_checkinode(dir, "post tracefs_mkdir-dir");

 out:
    unlock_dir(hidden_dir_dentry);
    if (!dentry->d_inode)
	d_drop(dentry);

    trace_mkdir_post(stoptr(dir->i_sb), dir, dentry, mode, err, &context);
    print_exit_status(err);
    return err;
}


STATIC int
tracefs_rmdir(inode_t *dir, dentry_t *dentry)
{
    int err = 0;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;
    void *context;

    print_entry_location();
    trace_rmdir_pre(stoptr(dir->i_sb), dir, dentry, &context);
    hidden_dentry = tracefs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    fist_checkinode(dir, "tracefs_rmdir-dir");

    dget(dentry);
    hidden_dir_dentry = lock_parent(hidden_dentry);

    /* avoid destroying the hidden inode if the file is in use */
    dget(hidden_dentry);
    err = vfs_rmdir(hidden_dir_dentry->d_inode, hidden_dentry);
    dput(hidden_dentry);

    if (!err)			/* vfs_rmdir does that */
	d_delete(hidden_dentry);

    // out_lock:
    fist_copy_attr_times(dir, hidden_dir_dentry->d_inode);
    /* copy the nlink count for our dentry and our parent's dentry */
    dir->i_nlink =  hidden_dir_dentry->d_inode->i_nlink;

    unlock_dir(hidden_dir_dentry);

    /*
     * call d_drop so the system "forgets" about us
     */
    if (!err)
	d_drop(dentry);

    dput(dentry);

    trace_rmdir_post(stoptr(dir->i_sb), dir, dentry, err, &context);
    print_exit_status(err);
    return err;
}


STATIC int
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
tracefs_mknod(inode_t *dir, dentry_t *dentry, int mode, int dev)
#else
tracefs_mknod(inode_t *dir,dentry_t *dentry, int mode, dev_t dev)
#endif
{
    int err;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;
    void *context;

    print_entry_location();
    trace_mknod_pre(stoptr(dir->i_sb), dir, dentry, mode, dev, &context);
    hidden_dentry = tracefs_hidden_dentry(dentry); /* CPW: Moved below print_entry_location */
    fist_checkinode(dir, "tracefs_mknod-dir");

    hidden_dir_dentry = lock_parent(hidden_dentry);

    err = vfs_mknod(hidden_dir_dentry->d_inode,
		    hidden_dentry,
		    mode,
		    dev);
    if (err || !hidden_dentry->d_inode)
	goto out;

    err = tracefs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (err)
	goto out;
    fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);

 out:
    unlock_dir(hidden_dir_dentry);
    if (!dentry->d_inode)
	d_drop(dentry);

    fist_checkinode(dir, "post tracefs_mknod-dir");
    trace_mknod_post(stoptr(dir->i_sb), dir, dentry, mode, dev, err, &context);
    print_exit_status(err);
    return err;
}


STATIC int
tracefs_rename(inode_t *old_dir, dentry_t *old_dentry,
	      inode_t *new_dir, dentry_t *new_dentry)
{
    int err;
    dentry_t *hidden_old_dentry;
    dentry_t *hidden_new_dentry;
    dentry_t *hidden_old_dir_dentry;
    dentry_t *hidden_new_dir_dentry;
    void *context;

    print_entry_location();
    trace_rename_pre(stoptr(old_dir->i_sb), old_dir, old_dentry, new_dir, new_dentry, &context);

    hidden_old_dentry = tracefs_hidden_dentry(old_dentry);/* CPW: Moved below print_entry_location */
    hidden_new_dentry = tracefs_hidden_dentry(new_dentry);

    fist_checkinode(old_dir, "tracefs_rename-old_dir");
    fist_checkinode(new_dir, "tracefs_rename-new_dir");

    dget(hidden_old_dentry);
    dget(hidden_new_dentry);
    hidden_old_dir_dentry = get_parent(hidden_old_dentry);
    hidden_new_dir_dentry = get_parent(hidden_new_dentry);
    //double_lock(hidden_old_dir_dentry, hidden_new_dir_dentry);
    lock_rename(hidden_old_dir_dentry, hidden_new_dir_dentry);

    err = vfs_rename(hidden_old_dir_dentry->d_inode, hidden_old_dentry,
		     hidden_new_dir_dentry->d_inode, hidden_new_dentry);
    if (err)
	goto out_lock;

    fist_copy_attr_all(new_dir, hidden_new_dir_dentry->d_inode);
    if (new_dir != old_dir)
	fist_copy_attr_all(old_dir, hidden_old_dir_dentry->d_inode);

 out_lock:
    // double_unlock will dput the new/old parent dentries whose refcnts
    // were incremented via dget above.
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    double_unlock(hidden_old_dir_dentry, hidden_new_dir_dentry);
#else
    //double_unlock(hidden_old_dir_dentry,hidden_new_dir_dentry);
    unlock_rename(hidden_old_dir_dentry, hidden_new_dir_dentry);
#endif
    dput(hidden_new_dentry);
    dput(hidden_old_dentry);

    fist_checkinode(new_dir, "post tracefs_rename-new_dir");
    trace_rename_post(stoptr(old_dir->i_sb), old_dir, old_dentry, new_dir, new_dentry, err, &context);
    print_exit_status(err);
    return err;
}


STATIC int
tracefs_readlink(dentry_t *dentry, char *buf, int bufsiz)
{
    int err;
    dentry_t *hidden_dentry;
    char *decoded_name,*hidden_buf;
    mm_segment_t old_fs;
    void *context;

    print_entry_location();
    trace_readlink_pre(stoptr(dentry->d_inode->i_sb), dentry, buf, bufsiz, &context);
    hidden_dentry = tracefs_hidden_dentry(dentry);/* CPW: Moved below print_entry_location */
    fist_print_dentry("tracefs_readlink IN", dentry);

    if (!hidden_dentry->d_inode->i_op ||
	!hidden_dentry->d_inode->i_op->readlink) {
	err = -EINVAL;
	goto out;
    }

    hidden_buf = KMALLOC(bufsiz,GFP_KERNEL);
    if(hidden_buf == NULL) {
	printk("Out of memory. \n");
	err = -ENOMEM;
	goto out;
    }

    old_fs = get_fs();
    set_fs(KERNEL_DS);
    err = hidden_dentry->d_inode->i_op->readlink(hidden_dentry,
						hidden_buf,
						bufsiz);
    set_fs(old_fs);
    if(err >= 0) {
	// don't do this, it's not zero-terminated!!!
	//	fist_dprint(7, "READLINK: link \"%s\", length %d\n", hidden_buf, err);
	err = tracefs_decode_filename(hidden_buf, err,
				      &decoded_name, DO_DOTS, dentry->d_inode, dentry->d_sb);
	if (err > 0) {
	    if (copy_to_user(buf, decoded_name, err))
		err = -EFAULT;
	    KFREE(decoded_name);
	}
	fist_copy_attr_atime(dentry->d_inode, hidden_dentry->d_inode);
    }
    KFREE(hidden_buf);

 out:
    trace_readlink_post(stoptr(dentry->d_inode->i_sb), dentry, buf, bufsiz, err, &context);
    print_exit_status(err);
    return err;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
STATIC int
#else
STATIC void *
#endif
tracefs_follow_link(dentry_t *dentry, struct nameidata *nd)
{
    char *buf;
    int len = PAGE_SIZE, err;
    mm_segment_t old_fs;
    void *context;

    print_entry_location();
    trace_follow_link_pre(stoptr(dentry->d_inode->i_sb), dentry, nd, &context);
    //    fist_print_dentry("tracefs_follow_link dentry IN", dentry);

    buf = KMALLOC(len, GFP_KERNEL);
    if (!buf) {
	err = -ENOMEM;
	goto out;
    }

    /* read the symlink, and then we will follow it */
    old_fs = get_fs();
    set_fs(KERNEL_DS);
    err = dentry->d_inode->i_op->readlink(dentry, buf, len);
    set_fs(old_fs);
    if (err < 0)
	goto out_free;

    buf[err] = 0;	// terminate the buffer -- XXX still needed?

    // XXX: FIXME w/ tracefs_encode_filename()
    /*
     * vfs_follow_link will increase the nd's mnt refcnt
     * we assume that some other VFS code decrements it.
     */
    err = vfs_follow_link(nd, buf);

 out_free:
    KFREE(buf);
 out:
#if 0
    if (err < 0) {
	dput(nd->dentry);
	printk("EZK follow_link() mnt_cnt %d\n", atomic_read(&nd->mnt->mnt_count));
	mntput(nd->mnt);
    }
#endif

    trace_follow_link_post(stoptr(dentry->d_inode->i_sb), dentry, nd, err, &context);
    print_exit_status(err);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
    return err;
#else
    return ERR_PTR(err);
#endif
}




STATIC int
tracefs_permission(inode_t *inode, int mask,struct nameidata* nd)
{
    inode_t *hidden_inode;
    int err;
    void *context;

    print_entry_location();
    trace_permission_pre(stoptr(inode->i_sb), inode, mask, &context);
    hidden_inode = itohi(inode);/* CPW: Moved below print_entry_location */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
    err = vfs_permission(inode, mask);
#else
    err = permission(hidden_inode, mask,nd);
#endif

    // out:
    trace_permission_post(stoptr(inode->i_sb),inode, mask, err, &context);
    print_exit_status(err);
    return err;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
STATIC int
tracefs_inode_revalidate(dentry_t *dentry)
{
    int err = 0;
    dentry_t *hidden_dentry;
    inode_t *hidden_inode;
    void *context;

    print_entry_location();
    if (dentry->d_inode)
        trace_inode_revalidate_pre(stoptr(dentry->d_inode->i_sb), dentry, &context);
    hidden_dentry = tracefs_hidden_dentry(dentry);/* CPW: Moved below print_entry_location */
    hidden_inode = hidden_dentry->d_inode;
    //    fist_print_dentry("tracefs_inode_revalidate IN", dentry);

    if (hidden_inode->i_op && hidden_inode->i_op->revalidate) {
	err = hidden_inode->i_op->revalidate(hidden_dentry);
	if (!err)
	    fist_copy_attr_all(dentry->d_inode, hidden_inode);
    }

    //    fist_print_dentry("tracefs_inode_revalidate OUT", dentry);
    if (dentry->d_inode)
        trace_inode_revalidate_post(stoptr(dentry->d_inode->i_sb), dentry, err, &context);
    print_exit_status(err);
    return err;
}
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)


STATIC int
tracefs_setattr(dentry_t *dentry, struct iattr *ia)
{
    int err = 0;
    dentry_t *hidden_dentry;
    inode_t *inode;
    inode_t *hidden_inode;
    void *context;

    print_entry_location();
    trace_setattr_pre(stoptr(dentry->d_inode->i_sb), dentry, ia, &context);
    hidden_dentry = tracefs_hidden_dentry(dentry);
    inode = dentry->d_inode;
    hidden_inode = itohi(inode);
    fist_checkinode(inode, "tracefs_setattr");

    if (mmap_support(dentry->d_sb)) {
	if (ia->ia_valid & ATTR_SIZE) {
	    err = tracefs_truncate(dentry, ia->ia_size);
	    if (err < 0)
		goto out;
	}
    }

    err = notify_change(hidden_dentry, ia);

out:
    /*
     * The lower file system might has changed the attributes, even if
     * notify_change above resulted in an error(!)  so we copy the
     * hidden_inode's attributes (and a few more) to our inode.
     */
    fist_copy_attr_all(inode, hidden_inode);

    fist_checkinode(inode, "post tracefs_setattr");
    trace_setattr_post(stoptr(dentry->d_inode->i_sb), dentry, ia, err, &context);
    print_exit_status(err);
    return err;
}


#ifdef NOT_USED_YET
STATIC int
tracefs_getattr(dentry_t *dentry, struct iattr *ia)
{
    return -ENOSYS;
}
#endif /* NOT_USED_YET */

#undef permission
struct inode_operations tracefs_symlink_iops =
{
    readlink:	tracefs_readlink,
    follow_link: tracefs_follow_link,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    permission:	tracefs_permission,
#else
    revalidate:	tracefs_inode_revalidate,
#endif
    setattr:	tracefs_setattr,
#if 0
    // XXX: off, b/c the VFS doesn't call getattr yet
    getattr:	tracefs_getattr,
#endif
};

struct inode_operations tracefs_dir_iops =
{
    create:	tracefs_create,
    lookup:	tracefs_lookup,
    link:	tracefs_link,
    unlink:	tracefs_unlink,
    symlink:	tracefs_symlink,
    mkdir:	tracefs_mkdir,
    rmdir:	tracefs_rmdir,
    mknod:	tracefs_mknod,
    rename:	tracefs_rename,
    /* no readlink/follow_link for non-symlinks */
    // off because we have setattr
    //    truncate:	tracefs_truncate,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    permission:	tracefs_permission,
#else
    revalidate:	tracefs_inode_revalidate,
#endif
    setattr:	tracefs_setattr,
#if 0
    // XXX: off, b/c the VFS doesn't call getattr yet
    getattr:	tracefs_getattr,
#endif
};

struct inode_operations tracefs_main_iops =
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    permission:	tracefs_permission,
#else
    revalidate:	tracefs_inode_revalidate,
#endif
    setattr:	tracefs_setattr,
#if 0
    // XXX: off, b/c the VFS doesn't call getattr yet
    getattr:	tracefs_getattr,
#endif
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
